__all__ = [
    'APIKeySecurityScheme',
    'AuthorizationCodeOAuthFlow',
    'Callback',
    'ClientCredentialsFlow',
    'Components',
    'Contact',
    'Discriminator',
    'Encoding',
    'Example',
    'ExternalDocumentation',
    'HTTPSecurityScheme',
    'Header',
    'ImplicitOAuthFlow',
    'Info',
    'License',
    'Link',
    'MediaType',
    'OAuth2SecurityScheme',
    'OAuthFlows',
    'OpenApiModel',
    'OpenIdConnectSecurityScheme',
    'Operation',
    'Parameter',
    'ParameterBase',
    'ParameterLocation',
    'PasswordOAuthFlow',
    'PathItem',
    'Paths',
    'Reference',
    'RequestBody',
    'Response',
    'Responses',
    'Schema',
    'SecurityRequirement',
    'SecuritySchemeBase',
    'Server',
    'ServerVariable',
    'Style',
    'Tag',
    'Type',
    'XML',
]

from .ext import *
from .model import *

type SchemaOrRef = Schema | Reference



import abc
import decimal as dec
import enum
import typing
from collections.abc import Mapping, Sequence
from enum import Enum
from typing import Annotated, Literal, Self, cast

import pydantic.alias_generators

from ...json_pointer import decode_json_pointer
from .base import (
    BaseModel,
    ExtendableModel,
    ModelWithAdditionalProperties,
    ModelWithPatternProperties,
    PropertyPattern,
    validate_example_xor_examples,
)
from .ext import LapidaryModelType


class Reference[Target](BaseModel):
    model_config = pydantic.ConfigDict(
        populate_by_name=True,
    )

    ref: typing.Annotated[str, pydantic.Field(alias='$ref')]


class Contact(ExtendableModel):
    name: str | None = None
    url: str | None = None
    email: str | None = None


class License(ExtendableModel):
    name: str
    url: str | None = None


class ServerVariable(ExtendableModel):
    enum: list[str] | None = None
    default: str
    description: str | None = None


class Type(Enum):
    array = 'array'
    boolean = 'boolean'
    integer = 'integer'
    number = 'number'
    object = 'object'
    string = 'string'


class Discriminator(BaseModel):
    property_name: Annotated[str, pydantic.Field(alias='propertyName')]
    mapping: dict[str, str] | None = None


class XML(ExtendableModel):
    name: str | None = None
    namespace: pydantic.AnyUrl | None = None
    prefix: str | None = None
    attribute: bool | None = False
    wrapped: bool | None = False


class Example(ExtendableModel):
    summary: str | None = None
    description: str | None = None
    value: typing.Any | None = None
    external_value: Annotated[str | None, pydantic.Field(alias='externalValue')] = None


class Style(Enum):
    deepObject = 'deepObject'
    form = 'form'
    label = 'label'
    matrix = 'matrix'
    pipeDelimited = 'pipeDelimited'
    simple = 'simple'
    spaceDelimited = 'spaceDelimited'


class SecurityRequirement(pydantic.RootModel):
    root: typing.Annotated[dict[str, list[str]], pydantic.Field(default_factory=dict)]


class ExternalDocumentation(ExtendableModel):
    description: str | None
    url: str


class ParameterLocation(enum.Enum):
    cookie = 'cookie'
    header = 'header'
    path = 'path'
    query = 'query'


class SecuritySchemeBase(abc.ABC, ExtendableModel):
    model_config = pydantic.ConfigDict(
        alias_generator=pydantic.alias_generators.to_camel,
    )
    type: str
    description: str | None = None


class APIKeySecurityScheme(SecuritySchemeBase):
    type: Literal['apiKey']
    name: str
    in_: typing.Annotated[ParameterLocation, pydantic.Field(alias='in')]

    format: Annotated[str, pydantic.Field(alias='x-lapidary-format')] = '{}'


class HTTPSecurityScheme(SecuritySchemeBase):
    type: Literal['http']
    scheme: str
    bearer_format: str | None = None


class OpenIdConnectSecurityScheme(SecuritySchemeBase):
    type: Literal['openIdConnect']
    open_id_connect_url: str


class ImplicitOAuthFlow(BaseModel):
    authorization_url: Annotated[str, pydantic.Field(alias='authorizationUrl')]
    refresh_url: Annotated[str | None, pydantic.Field(alias='refreshUrl')] = None
    scopes: dict[str, str]


class PasswordOAuthFlow(BaseModel):
    token_url: Annotated[str, pydantic.Field(alias='tokenUrl')]
    refresh_url: Annotated[str | None, pydantic.Field(alias='refreshUrl')] = None
    scopes: dict[str, str] | None = None


class ClientCredentialsFlow(PasswordOAuthFlow):
    pass


class AuthorizationCodeOAuthFlow(BaseModel):
    authorization_url: Annotated[str, pydantic.Field(alias='authorizationUrl')]
    token_url: Annotated[str, pydantic.Field(alias='tokenUrl')]
    refresh_url: Annotated[str | None, pydantic.Field(alias='refreshUrl')] = None
    scopes: dict[str, str] | None = None


class Info(ExtendableModel):
    title: str
    description: str | None = None
    terms_of_service: Annotated[str | None, pydantic.Field(alias='termsOfService')] = None
    contact: Contact | None = None
    license: License | None = None
    version: str


class Server(ExtendableModel):
    url: str
    description: str | None = None
    variables: dict[str, ServerVariable] | None = None


def validate_list_unique(v: Sequence[typing.Any]) -> Sequence[typing.Any]:
    if len(set(v)) != len(v):
        raise ValueError('not unique')
    return v


UniqueListValidator = pydantic.AfterValidator(validate_list_unique)


class Schema(ExtendableModel):
    title: str | None = None
    type: Type | None = None

    # type == number or type == integer
    multiple_of: typing.Annotated[dec.Decimal | None, pydantic.Field(alias='multipleOf', gt=0.0)] = None
    maximum: dec.Decimal | None = None
    exclusive_maximum: Annotated[bool | None, pydantic.Field(alias='exclusiveMaximum')] = False
    minimum: dec.Decimal | None = None
    exclusive_minimum: Annotated[bool | None, pydantic.Field(alias='exclusiveMinimum')] = False

    # type == string
    max_length: typing.Annotated[int | None, pydantic.Field(alias='maxLength', ge=0)] = None
    min_length: typing.Annotated[int, pydantic.Field(alias='minLength', ge=0)] = 0
    pattern: str | None = None

    # type == array
    items: 'None | Reference[Schema] | Schema' = None
    max_items: typing.Annotated[int | None, pydantic.Field(alias='maxItems', ge=0)] = None
    min_Items: typing.Annotated[int | None, pydantic.Field(alias='minItems', ge=0)] = 0
    unique_items: Annotated[bool | None, pydantic.Field(alias='uniqueItems')] = False

    # type == object
    max_properties: typing.Annotated[int | None, pydantic.Field(alias='maxProperties', ge=0)] = None
    min_properties: typing.Annotated[int | None, pydantic.Field(alias='minProperties', ge=0)] = 0
    required: typing.Annotated[list[str], pydantic.Field(min_length=1, default_factory=list), UniqueListValidator]
    properties: 'typing.Annotated[dict[str, Reference[Schema] | Schema], pydantic.Field(default_factory=dict)]'
    additional_properties: 'Annotated[bool | Reference[Schema] | Schema, pydantic.Field(alias="additionalProperties")]' = True

    # type == string or type = number or type == integer
    format: str | None = None

    enum: typing.Annotated[
        list | None,
        pydantic.Field(min_length=1),
    ] = None

    not_: 'Annotated[None | Reference[Schema] | Schema, pydantic.Field(alias="not")]' = None
    all_of: 'Annotated[None | list[Reference[Schema] | Schema], pydantic.Field(alias="allOf")]' = None
    one_of: 'Annotated[None | list[Reference[Schema] | Schema], pydantic.Field(alias="oneOf")]' = None
    any_of: 'Annotated[None | list[Reference[Schema] | Schema], pydantic.Field(alias="anyOf")]' = None

    description: str | None = None
    default: typing.Any = None
    nullable: bool = False
    discriminator: Discriminator | None = None
    read_only: Annotated[bool, pydantic.Field(alias='readOnly')] = False
    write_only: Annotated[bool, pydantic.Field(alias='writeOnly')] = False
    example: typing.Any = None
    external_docs: Annotated[ExternalDocumentation | None, pydantic.Field(alias='externalDocs')] = None
    deprecated: bool = False
    xml: XML | None = None

    lapidary_names: typing.Annotated[
        dict[str | None, typing.Any] | None,
        pydantic.Field(
            alias='x-lapidary-names',
            default_factory=dict,
            description='Mapping of keys used in the JSON document and variable names in the generated Python code. '
            'Applicable to enum values or object properties.',
        ),
    ]
    lapidary_name: typing.Annotated[str | None, pydantic.Field(alias='x-lapidary-type-name')] = None
    lapidary_model_type: typing.Annotated[LapidaryModelType | None, pydantic.Field(alias='x-lapidary-modelType')] = None


class Tag(ExtendableModel):
    name: str
    description: str | None = None
    external_docs: Annotated[ExternalDocumentation | None, pydantic.Field(alias='externalDocs')] = None


class Link(ExtendableModel):
    operation_id: Annotated[str | None, pydantic.Field(alias='operationId')] = None
    operation_ref: Annotated[str | None, pydantic.Field(alias='operationRef')] = None
    parameters: dict[str, typing.Any] | None = None
    request_body: Annotated[typing.Any | None, pydantic.Field(alias='requestBody')] = None
    description: str | None = None
    server: Server | None = None


class OAuthFlows(ExtendableModel):
    implicit: ImplicitOAuthFlow | None = None
    password: PasswordOAuthFlow | None = None
    client_credentials: Annotated[ClientCredentialsFlow | None, pydantic.Field(alias='clientCredentials')] = None
    authorization_code: Annotated[AuthorizationCodeOAuthFlow | None, pydantic.Field(alias='authorizationCode')] = None


class OAuth2SecurityScheme(SecuritySchemeBase):
    type: Literal['oauth2']
    flows: OAuthFlows


SecurityScheme: typing.TypeAlias = (
    APIKeySecurityScheme | HTTPSecurityScheme | OAuth2SecurityScheme | OpenIdConnectSecurityScheme
)


class ParameterBase(ExtendableModel):
    in_: typing.Annotated[ParameterLocation, pydantic.Field(alias='in')]
    description: str | None = None
    required: bool = False
    deprecated: bool = False
    allow_empty_value: Annotated[bool, pydantic.Field(alias='allowEmptyValue')] = False

    content: 'typing.Annotated[dict[str, MediaType] | None, pydantic.Field(max_length=1, min_length=1)]' = None

    style: Style | None = None
    explode: bool | None = None
    allow_reserved: Annotated[bool | None, pydantic.Field(alias='allowReserved')] = False
    schema_: typing.Annotated[None | Reference[Schema] | Schema, pydantic.Field(alias='schema')] = None
    example: typing.Any = None
    examples: dict[str, Reference[Example] | Example] | None = None

    @pydantic.model_validator(mode='before')
    @staticmethod
    def _validate(values: Mapping[str, typing.Any]):
        if not isinstance(values, Mapping | MediaType):
            raise TypeError(type(values))
        if 'content' not in values != 'schema' not in values:
            raise ValueError('content or schema required')
        validate_example_xor_examples(values)
        return values


class Header(ParameterBase):
    in_: typing.Annotated[ParameterLocation, pydantic.Field(alias='in')] = ParameterLocation.header
    style: Style = Style.simple


class Encoding(ExtendableModel):
    content_type: Annotated[str | None, pydantic.Field(alias='contentType')] = None
    headers: dict[str, Header] | None = None
    style: Style | None = None
    explode: bool | None = None
    allow_reserved: Annotated[bool | None, pydantic.Field(alias='allowReserved')] = False


class MediaType(ExtendableModel):
    schema_: typing.Annotated[None | Reference[Schema] | Schema, pydantic.Field(alias='schema')] = None
    example: typing.Any | None = None
    examples: dict[str, Reference[Example] | Example] | None = None
    encoding: dict[str, Encoding] | None = None

    @pydantic.model_validator(mode='before')
    @staticmethod
    def _validate(values: Mapping[str, typing.Any]):
        if not isinstance(values, Mapping | MediaType):
            raise TypeError(type(values))
        validate_example_xor_examples(values)
        return values


class Response(ExtendableModel):
    description: str
    headers: Annotated[dict[str, Reference[Header] | Header], pydantic.Field(default_factory=dict)]
    content: 'typing.Annotated[dict[str, MediaType], pydantic.Field(default_factory=dict)]'
    links: dict[str, Reference[Link] | Link] | None = None


class Responses(ExtendableModel, ModelWithPatternProperties):
    responses: typing.Annotated[
        dict[str, Reference[Response] | Response],
        pydantic.Field(default_factory=dict, min_length=1),
        PropertyPattern(r'^[1-5](?:\d{2}|XX)|default$'),
    ]


class Parameter(ParameterBase):
    name: str

    lapidary_name: typing.Annotated[str | None, pydantic.Field(alias='x-lapidary-name')] = None

    @property
    def effective_name(self) -> str:
        return self.lapidary_name or self.name

    def __hash__(self) -> int:
        return (hash(self.name) << 2) + hash(self.in_)


class RequestBody(ExtendableModel):
    description: str | None = None
    content: 'dict[str, MediaType]'
    required: bool | None = False


class Operation(ExtendableModel):
    tags: list[str] | None = None
    summary: str | None = None
    description: str | None = None
    external_docs: Annotated[ExternalDocumentation | None, pydantic.Field(alias='externalDocs')] = None
    operation_id: Annotated[str | None, pydantic.Field(alias='operationId')] = None
    parameters: typing.Annotated[
        list[Reference[Parameter] | Parameter], UniqueListValidator, pydantic.Field(default_factory=list)
    ]
    request_body: Annotated[None | Reference[RequestBody] | RequestBody, pydantic.Field(alias='requestBody')] = None
    responses: Responses
    callbacks: 'dict[str, Reference[Callback] | Callback] | None' = None
    deprecated: bool | None = False
    security: list[SecurityRequirement] | None = None
    servers: list[Server] | None = None


class PathItem(ModelWithAdditionalProperties):
    summary: str | None = None
    description: str | None = None
    servers: list[Server] | None = None
    parameters: typing.Annotated[
        list[Reference[Parameter] | Parameter], UniqueListValidator, pydantic.Field(default_factory=list)
    ]
    __pydantic_extra__: dict[str, Operation]
    """Keys are HTTP methods"""


class Paths(ModelWithPatternProperties):
    paths: typing.Annotated[dict[str, PathItem], pydantic.Field(default_factory=dict), PropertyPattern('^/')]


class Callback(ModelWithAdditionalProperties):
    __pydantic_extra__ = dict[str, Reference[PathItem] | PathItem]


class Components(ExtendableModel):
    schemas: Annotated[dict[str, Reference[Schema] | Schema], pydantic.Field(default_factory=dict)]
    responses: dict[str, Reference[Response] | Response] | None = None
    parameters: dict[str, Reference[Parameter] | Parameter] | None = None
    examples: dict[str, Reference[Example] | Example] | None = None
    request_bodies: Annotated[
        dict[str, Reference[RequestBody] | RequestBody] | None, pydantic.Field(alias='requestBodies')
    ] = None
    headers: dict[str, Reference[Header] | Header] | None = None
    security_schemes: Annotated[
        dict[str, Reference[SecurityScheme] | SecurityScheme] | None, pydantic.Field(alias='securitySchemes')
    ] = None
    links: dict[str, Reference[Link] | Link] | None = None
    callbacks: dict[str, Reference[Callback] | Callback] | None = None


class OpenApiModel(ExtendableModel):
    """
    Validation schema for OpenAPI Specification 3.0.X.
    """

    model_config = pydantic.ConfigDict(
        extra='forbid',
        populate_by_name=True,
    )

    openapi: typing.Annotated[str, pydantic.Field(pattern='^3\\.0\\.\\d(-.+)?$')]
    info: Info
    external_docs: Annotated[ExternalDocumentation | None, pydantic.Field(alias='externalDocs')] = None
    servers: list[Server] = (Server(url='/'),)
    security: list[SecurityRequirement] | None = None
    tags: typing.Annotated[list[Tag] | None, UniqueListValidator] = None
    paths: Paths
    components: Components | None = None

    lapidary_headers_global: typing.Annotated[
        dict[str, Header],
        pydantic.Field(
            alias='x-lapidary-headers-global',
            default_factory=dict,
        ),
    ]
    """Headers added to every request.
    Unlike with operation headers, the default value found in the schema is sent over the wire"""

    lapidary_responses_global: typing.Annotated[
        Responses | None,
        pydantic.Field(
            alias='x-lapidary-responses-global',
            description='Common Responses, added to every operation. '
            'Values in Responses declared in Operations override values in this one.',
            default=None,
        ),
    ]

    def resolve_ref[Target](self, ref: Reference[Target]) -> tuple[Target, str]:
        pointer = ref.ref
        target = self._resolve_ref(pointer)
        stack = [pointer]

        while isinstance(target, Reference):
            pointer = target.ref
            if pointer in stack:
                raise ValueError('Circular references', stack, pointer)
            else:
                stack.append(pointer)
            target = self._resolve_ref(pointer)
        return cast(Target, target), pointer

    def _resolve_ref[Target](self: Self, ref_str: str) -> Target | Reference[Target]:
        """Resolve ref without recursion"""
        obj: typing.Any = self
        for name in ref_str.split('/')[1:]:
            name = decode_json_pointer(name)
            obj = _resolve_name(obj, name)
        return obj


def _resolve_name(src: typing.Any, name: str) -> typing.Any:
    if isinstance(src, Paths) and name in src.paths:
        src = src.paths
    elif isinstance(src, PathItem) and name in src.model_extra:
        src = src.model_extra

    if isinstance(src, Sequence):
        return src[int(name)]
    elif isinstance(src, Mapping):
        return src[name]
    else:
        if hasattr(src, name):
            return getattr(src, name)
        for field_name, field_info in cast(pydantic.BaseModel, src).model_fields.items():
            if name == field_info.alias:
                return getattr(src, field_name)
        else:
            raise AttributeError(name)



import typing
from collections.abc import Iterable
from pathlib import PurePath
from typing import Self


class ModulePath:
    _SEP = '.'

    def __init__(self, module: str | Iterable[str], is_module: bool = True):
        if isinstance(module, str):
            module = module.strip()
            if module.strip() != module:
                raise ValueError(f'"{module}"')
            parts: Iterable[str] = module.split(ModulePath._SEP)
        else:
            parts = module

        if isinstance(parts, list):
            if len(parts) == 0:
                raise ValueError(module)
            self.parts = parts
        else:
            raise ValueError(module)

        self._is_module = is_module

    def to_path(self, root: PurePath | None = None):
        parts = list(self.parts)
        if not self._is_module:
            parts.append('__init__')
        return (root or PurePath()).joinpath(*parts)

    def parent(self) -> typing.Self | None:
        if len(self.parts) == 1:
            return None
        return ModulePath(self.parts[:-1], False)

    def __truediv__(self, other: str | Iterable[str]):
        if isinstance(other, str):
            other = [other]
        return ModulePath([*self.parts, *other])

    def __repr__(self):
        return ModulePath._SEP.join(self.parts)

    def __eq__(self, other: object):
        if not isinstance(other, ModulePath):
            return NotImplemented
        return self.parts == other.parts

    def __hash__(self) -> int:
        return hash(self.__str__())

    def __matmul__(self, other):
        if not isinstance(other, ModulePath):
            return NotImplemented

        if self.parts[: len(other.parts)] != other.parts:
            raise ValueError('Not related')

        return ModulePath(self.parts[len(other.parts) :])



from typing import Self


class Stack:
    __slots__ = ('path',)

    def __init__(self, stack=('#',)) -> None:
        self.path: tuple[str, ...] = stack

    @classmethod
    def from_str(cls, pointer: str) -> Self:
        return cls(tuple(pointer.split('/')))

    def __repr__(self):
        return '/'.join(self.path)

    def push(self, *names: str) -> Self:
        return Stack(self.path + names)

    def top(self) -> str:
        return self.path[-1]

    def __hash__(self) -> int:
        return hash(self.path)

    def __eq__(self, other):
        return isinstance(other, Stack) and self.path == other.path

    def __getitem__(self, item: int) -> str:
        return self.path[item]



import tomllib

import anyio
import pydantic

PYPROJ_TOML = 'pyproject.toml'


class Config(pydantic.BaseModel):
    document_path: str | None = None
    origin: pydantic.AnyHttpUrl | None = None
    package: str
    patches: str = 'src/patches'


async def load_config(project_root: anyio.Path) -> Config:
    text = await (project_root / PYPROJ_TOML).read_text()
    pyproj = tomllib.loads(text)
    pyproj_dict = pyproj['tool']['lapidary']
    return Config.model_validate(pyproj_dict)



import abc
import logging
from collections.abc import AsyncIterator, Mapping, Sequence
from typing import cast

import anyio
import httpx
import ruamel.yaml
from jsonpatch import JsonPatch

from .config import Config

logger = logging.getLogger(__name__)


async def load_document(root: anyio.Path, config: Config) -> Mapping:
    logger.info('Load OpenAPI document')
    spec_dict: dict = cast(dict, await load_parse(root, config.document_path))

    from jsonpointer import JsonPointerException

    async for path, patch in load_patches(root / config.patches):
        try:
            spec_dict = patch.apply(spec_dict)
        except JsonPointerException as error:
            raise JsonPointerException(f'Error while applying patch {str(path)}', error.args[0][:100] + '...') from None

    return spec_dict


async def load_parse(root: anyio.Path, path: str | anyio.Path) -> Mapping | Sequence:
    text = await document_handler_for(root, path).load()
    return parse(text)


def parse(text: str) -> Mapping:
    yaml = ruamel.yaml.YAML(typ='safe')
    return yaml.load(text)


async def load_patches(patches_root: anyio.Path) -> AsyncIterator[tuple[anyio.Path, JsonPatch]]:
    patches = sorted([p async for p in patches_root.rglob('*[yamljson]')])

    if not patches:
        return

    logger.info('Load patches')

    for patch in patches:
        if patch.suffix in ('.yaml', '.yml', '.json'):
            path = patch.relative_to(patches_root)
            yield path, JsonPatch(await load_parse(patches_root, path))


class DocumentHandler[P: str | anyio.Path](abc.ABC):
    def __init__(self, path: P) -> None:
        self._path = path

    @abc.abstractmethod
    async def load(self) -> str:
        """Return document text and, which could be URL path or local file path."""

    @property
    @abc.abstractmethod
    def is_url(self) -> bool:
        pass

    @property
    def path(self) -> str:
        return str(self._path)

    @abc.abstractmethod
    async def save_to(self, target: anyio.Path) -> str:
        """Save the document.
        :param target: target directory
        :return: file name of the saved document"""


class FileDocumentHandler(DocumentHandler[anyio.Path]):
    def __init__(self, project_root: anyio.Path, path: anyio.Path) -> None:
        assert isinstance(project_root, anyio.Path)
        assert isinstance(path, anyio.Path)
        super().__init__(path)
        self._project_root = project_root

    async def load(self) -> str:
        return await (self._project_root / self._path).read_text()

    async def save_to(self, target: anyio.Path) -> str:
        import shutil

        shutil.copyfile(self._path, target / self._path.name)
        return self._path.name

    @property
    def is_url(self) -> bool:
        return False


class HttpDocumentHandler(DocumentHandler[str]):
    def __init__(self, path: str) -> None:
        super().__init__(path)
        self._client = httpx.AsyncClient()
        self._cache: str | None = None

    async def load(self) -> str:
        if not self._cache:
            response = await self._client.get(self._path)
            self._cache = response.text
        assert self._cache is not None
        return self._cache

    def _file_name(self) -> str:
        from os.path import split
        from urllib.parse import urlparse

        parsed = urlparse(self._path)
        _, name = split(parsed.path)
        return name

    async def save_to(self, target: anyio.Path) -> str:
        text = await self.load()
        file_name = self._file_name()
        await (target / file_name).write_text(text)
        return file_name

    @property
    def is_url(self) -> bool:
        return True


def document_handler_for(document_root: anyio.Path, path: str | anyio.Path) -> DocumentHandler:
    match path:
        case str() if path.startswith(('http://', 'https://')):
            return HttpDocumentHandler(path)
        case _:
            return FileDocumentHandler(document_root, anyio.Path(path))



import enum

__all__ = [
    'LapidaryModelType',
]


class LapidaryModelType(enum.Enum):
    model = 'model'
    """The default type, rendered as a subclass of pydantic.BaseModel."""

    exception = 'exception'
    """Error typy, not used as return type, if received, it's raised."""



import functools
import logging
from collections.abc import Callable, Iterable, Mapping, MutableMapping
from typing import Any, cast

from mimeparse import parse_media_range

from .. import json_pointer, names
from . import openapi, python
from .refs import resolve_ref
from .schema import OpenApi30SchemaConverter, resolve_type_hint
from .stack import Stack

logger = logging.getLogger(__name__)


class OpenApi30Converter:
    def __init__(
        self,
        root_package: python.ModulePath,
        source: openapi.OpenApiModel,
        origin: str | None,
        schema_converter: OpenApi30SchemaConverter | None = None,
        path_progress: Callable[[Any], None] | None = None,
    ):
        self.root_package = root_package
        self.global_headers: dict[str, python.Parameter] = {}
        self.src = source
        self._origin = origin
        self._path_progress = path_progress

        self.schema_converter = schema_converter or OpenApi30SchemaConverter(self.root_package, self.resolve_ref)

        self.target = python.ClientModel(
            client=python.ClientModule(
                path=root_package,
                body=python.ClientClass(init_method=python.ClientInit()),
            ),
            package=str(root_package),
        )

        self._response_mime_maps: MutableMapping[Stack, python.MimeMap] = {}

    def process(self) -> python.ClientModel:
        stack = Stack()

        map_process(
            self.src,
            stack,
            {
                'servers': self.process_servers,
                'lapidary_responses_global': self.process_global_responses,
                'lapidary_headers_global': self.process_global_headers,
                'security': self.process_global_security,
                'paths': self.process_paths,
            },
        )

        self.target.schemas.extend(self.schema_converter.schema_modules)
        return self.target

    def process_servers(self, value: list[openapi.Server] | None, stack: Stack) -> None:
        logger.debug('Process servers %s', stack)

        if not value:
            logger.warning('No servers found')
            return
        if len(value) > 1:
            logger.warning('Multiple servers found, using the first')

        server = value[0]
        server_url = server.url
        if server.variables:
            server_url = server_url.format(**{name: var.default for name, var in server.variables.items()})

        if self._origin:
            from urllib.parse import urljoin

            server_url = urljoin(self._origin, server_url)

        self.target.client.body.init_method.base_url = server_url

    def process_global_headers(self, value: Mapping[str, openapi.Header], stack: Stack) -> None:
        logger.debug('Process global headers %s', stack)
        if not value:
            return

        for header_name, header in value.items():
            self.global_headers[header_name] = self.process_parameter(header, stack.push(header_name))

    def process_global_responses(self, value: openapi.Responses | None, stack: Stack) -> None:
        logger.debug('Process global responses %s', stack)
        if not value:
            return

        self.global_responses = {
            code: self.process_response(response, stack.push(code)) for code, response in value.responses.items()
        }

    # Not providing process_parameters (plural) as each caller calls it in a different context
    # (list, map or map with defaults)

    def _process_schema_or_content(
        self,
        value: openapi.ParameterBase,
        stack: Stack,
    ) -> tuple[python.TypeHint, str]:
        if value.schema_ and value.content:
            raise ValueError()
        if value.schema_:
            media_type: str | None = None
            # encoding: str | None = None
            return self.schema_converter.process_schema(value.schema_, stack.push('schema'), value.required), media_type
        elif value.content:
            media_type, media_type_obj = next(iter(value.content.items()))
            # encoding = media_type_obj.encoding
            return self.schema_converter.process_schema(
                media_type_obj.schema_, stack.push('content', media_type), value.required
            ), media_type
        else:
            raise TypeError(f'{stack}: schema or content is required')

    @resolve_ref
    def process_parameter(self, value: openapi.Parameter, stack: Stack) -> python.Parameter:
        logger.debug('process_parameter %s', stack)

        if not isinstance(value, openapi.ParameterBase):
            raise TypeError(f'Expected Parameter object at {stack}, got {type(value).__name__}.')

        typ, media_type = self._process_schema_or_content(value, stack)

        return python.Parameter(
            name=parameter_name(value),
            alias=value.name,
            type=typ,
            in_=value.in_,
            style=value.style,
            explode=value.explode,
            required=value.required,
            media_type=media_type,
        )

    def process_paths(self, value: openapi.Paths, stack: Stack) -> None:
        for path, path_item in value.paths.items():
            self.process_path(path_item, stack.push(json_pointer.encode_json_pointer(path)))

    def process_path(
        self,
        value: openapi.PathItem,
        stack: Stack,
    ) -> None:
        common_params_stack = stack.push('parameters')
        common_params = [
            self.process_parameter(param, common_params_stack.push(str(idx)))
            for idx, param in enumerate(value.parameters)
        ]

        for method, operation in value.model_extra.items():
            self.process_operation(operation, stack.push(method), common_params)
        self._path_progress(json_pointer.decode_json_pointer(stack.top()))

    @resolve_ref
    def process_request_body(self, value: openapi.RequestBody, stack: Stack) -> python.MimeMap:
        # TODO handle required
        return self.process_content(value.content, stack.push('content'))

    def process_responses(self, value: openapi.Responses, stack: Stack) -> python.ResponseMap:
        return {code: self.process_response(response, stack.push(code)) for code, response in value.responses.items()}

    @resolve_ref
    def process_response(
        self,
        value: openapi.Response,
        stack: Stack,
    ) -> python.MimeMap:
        assert isinstance(value, openapi.Response)

        if stack in self._response_mime_maps:
            return self._response_mime_maps[stack]

        headers_stack = stack.push('headers')

        headers = [self.process_header(header, headers_stack.push(name)) for name, header in value.headers.items()]

        mime_map_body_only = self.process_content(value.content, stack.push('content'))
        mime_map = {
            mime_type: self._create_envelope_model(
                body_type=body_type,
                headers=headers,
                stack=stack,
            )
            for mime_type, body_type in mime_map_body_only.items()
        }
        self._response_mime_maps[stack] = mime_map
        return mime_map

    @resolve_ref
    def process_header(self, value: openapi.Header, stack: Stack) -> python.ResponseHeader:
        alias = stack.top()

        typ, _ = self._process_schema_or_content(value, stack)

        return python.ResponseHeader(name=names.maybe_mangle_name(alias), alias=alias, type=typ, annotation='Header')

    def _create_envelope_model(
        self,
        body_type: python.TypeHint,
        headers: Iterable[python.ResponseHeader],
        stack: Stack,
    ) -> python.TypeHint:
        envelope = python.ResponseEnvelopeModel(
            name='Response',
            headers=headers,
            body_type=body_type,
        )
        type_hint = resolve_type_hint(str(self.root_package), stack.push('response', 'Response'))
        self.target.add_response_envelope_module(
            python.ResponseEnvelopeModule(
                path=python.ModulePath(type_hint.module, is_module=True),
                body=envelope,
            )
        )
        return type_hint

    def process_content(self, value: Mapping[str, openapi.MediaType], stack: Stack) -> python.MimeMap:
        """Returns: {mime_type: response body type hint}"""
        types = {}
        for mime, media_type in value.items():
            mime_parsed = parse_media_range(mime)
            if mime_parsed[:2] != ('application', 'json'):
                continue
            types[mime] = self.schema_converter.process_schema(media_type.schema_, stack.push(mime, 'schema'))
        return types

    def process_operation(
        self,
        value: openapi.Operation,
        stack: Stack,
        common_params: Iterable[python.Parameter],
    ) -> None:
        logger.debug('Process operation %s', stack)

        if not value.operation_id:
            raise ValueError(f'{stack}: operationId is required')

        params = self._mk_params(value.parameters, stack.push('parameters'), common_params)

        request_body = (
            self.process_request_body(value.request_body, stack.push('requestBody')) if value.request_body else {}
        )
        responses = self.process_responses(value.responses, stack.push('responses'))
        security = self.process_security(value.security, stack.push('security'))

        model = python.OperationFunction(
            name=value.operation_id,
            method=stack.top(),
            path=json_pointer.decode_json_pointer(stack[-2]),
            request_body=request_body,
            params=params,
            responses=responses,
            security=security,
        )

        self.target.client.body.methods.append(model)

    def _mk_params(
        self,
        value: list[openapi.Parameter | openapi.Reference[openapi.Parameter]],
        stack: Stack,
        common_params: Iterable[python.Parameter],
    ) -> Iterable[python.Parameter]:
        params = {}
        for param in common_params:
            params[param.name] = param
        for idx, oa_param in enumerate(value):
            param = self.process_parameter(oa_param, stack.push(str(idx)))
            params[param.name] = param
        return list(params.values())

    def process_global_security(self, value: Iterable[openapi.SecurityRequirement] | None, stack: Stack) -> None:
        self.target.client.body.init_method.security = self.process_security(value, stack)

    def process_security(
        self, value: Iterable[openapi.SecurityRequirement] | None, stack: Stack
    ) -> python.SecurityRequirements | None:
        logger.debug('Process security %s', stack)
        if value is None:
            return None

        return [self.process_security_requirement(item, stack.push(str(idx))) for idx, item in enumerate(value)]

    def process_security_requirement(
        self, value: openapi.SecurityRequirement, stack: Stack
    ) -> Mapping[str, Iterable[str]]:
        logger.debug('Process security requirement %s', stack)
        schemes_root = Stack(('#', 'components', 'securitySchemes'))
        for scheme_name, scopes in value.root.items():
            scheme_stack = schemes_root.push(scheme_name)
            self.process_security_scheme(
                openapi.Reference[openapi.SecuritySchemeBase](ref=str(scheme_stack)), scheme_stack
            )
        return value.root

    # need separate method to resolve references before calling a single-dispatched method
    @resolve_ref
    def process_security_scheme(self, value: openapi.SecuritySchemeBase, stack: Stack) -> None:
        self.process_security_scheme_(value, stack)

    @functools.singledispatchmethod
    def process_security_scheme_(self, _: openapi.SecuritySchemeBase, stack: Stack) -> None:
        raise NotImplementedError(str(stack))

    @process_security_scheme_.register(openapi.APIKeySecurityScheme)
    def _(self, value: openapi.APIKeySecurityScheme, stack: Stack) -> None:
        logger.debug('Process API key security scheme %s', stack)
        auth_name = stack.top()
        flow_name = f'api_key_{auth_name}'
        if flow_name in self.target.security_schemes:
            return

        self.target.security_schemes[flow_name] = python.ApiKeyAuth(
            name=auth_name,
            python_name=names.maybe_mangle_name(auth_name),
            key=value.name,
            location=value.in_,
            format=value.format,
        )

    @process_security_scheme_.register(openapi.OAuth2SecurityScheme)
    def _(self, value: openapi.OAuth2SecurityScheme, stack: Stack) -> None:
        auth_name = stack.top()
        if auth_name in self.target.security_schemes:
            return

        logger.debug('Process OAuth2 security scheme %s', stack)
        map_process(
            value.flows,
            stack.push('flows'),
            {
                'implicit': self.process_oauth2_implicit,
                'password': self.process_oauth2_password,
                'authorization_code': self.process_oauth2_auth_code,
                'client_credentials': self.process_oauth2_client_credentials,
            },
            True,
        )

    def process_oauth2_implicit(self, value: openapi.ImplicitOAuthFlow, stack: Stack) -> None:
        if value.refresh_url:
            raise NotImplementedError(stack.push('refreshUrl'))

        auth_name = stack[-3]
        self.target.security_schemes[f'oauth2_implicit_{auth_name}'] = python.ImplicitOAuth2Flow(
            name=auth_name,
            python_name=names.maybe_mangle_name(auth_name),
            authorization_url=value.authorization_url,
            scopes=value.scopes,
        )

    def process_oauth2_password(self, value: openapi.PasswordOAuthFlow, stack: Stack) -> None:
        if value.refresh_url:
            raise NotImplementedError(stack.push('refreshUrl'))

        auth_name = stack[-3]
        self.target.security_schemes[f'oauth2_password_{auth_name}'] = python.PasswordOAuth2Flow(
            name=auth_name,
            python_name=names.maybe_mangle_name(auth_name),
            token_url=value.token_url,
            scopes=value.scopes,
        )

    def process_oauth2_auth_code(self, value: openapi.AuthorizationCodeOAuthFlow, stack: Stack) -> None:
        if value.refresh_url:
            raise NotImplementedError(stack.push('refreshUrl'))

        auth_name = stack[-3]
        self.target.security_schemes[f'oauth2_auth_code_{auth_name}'] = python.AuthorizationCodeOAuth2Flow(
            name=auth_name,
            python_name=names.maybe_mangle_name(auth_name),
            authorization_url=value.authorization_url,
            token_url=value.token_url,
            scopes=value.scopes,
        )

    def process_oauth2_client_credentials(self, value: openapi.ClientCredentialsFlow, stack: Stack) -> None:
        if value.refresh_url:
            raise NotImplementedError(stack.push('refreshUrl'))

        auth_name = stack[-3]
        self.target.security_schemes[f'oauth2_client_credentials_{auth_name}'] = python.ClientCredentialsOAuth2Flow(
            name=auth_name,
            python_name=names.maybe_mangle_name(auth_name),
            token_url=value.token_url,
            scopes=value.scopes,
        )

    @process_security_scheme_.register(openapi.HTTPSecurityScheme)
    def _(self, value: openapi.HTTPSecurityScheme, stack: Stack) -> None:
        logger.debug('Process HTTP security scheme %s', stack)
        auth_name = stack.top()
        flow_name = f'http_{auth_name}'

        if flow_name in self.target.security_schemes:
            return

        try:
            self.target.security_schemes[flow_name] = HTTP_SCHEMES[value.scheme.lower()](
                name=auth_name,
                python_name=names.maybe_mangle_name(auth_name),
            )
        except KeyError:
            raise NotImplementedError(stack.push('scheme'), value.scheme) from None

    def resolve_ref[Target](self, ref: openapi.Reference[Target]) -> tuple[Target, Stack]:
        """Resolve reference to OpenAPI object and its direct path."""
        value, pointer = self.src.resolve_ref(ref)
        return cast(Target, value), Stack.from_str(pointer)


def parameter_name(value: openapi.Parameter) -> str:
    return value.lapidary_name or names.maybe_mangle_name(value.name) + '_' + value.in_.name[0]


def map_process(
    obj: Any, stack: Stack, processors: Mapping[str, Callable[[Any, Stack], Any]], warn: bool = False
) -> None:
    for key, process in processors.items():
        try:
            if value := getattr(obj, key, None):
                alias = type(obj).model_fields[key].alias
                substack = stack.push(alias or key)
                process(value, substack)
        except NotImplementedError as e:
            if warn:
                logger.warning('Not implemented: %s', e.args[0])
            else:
                raise


HTTP_SCHEMES = {
    'basic': python.HttpBasicAuth,
    'digest': python.HttpDigestAuth,
}



import re
import typing
from collections.abc import Mapping

import pydantic

from ...pydantic_utils import find_annotation_optional


class BaseModel(pydantic.BaseModel):
    model_config = pydantic.ConfigDict(
        populate_by_name=True,
        extra='forbid',
        frozen=True,
    )


class ExtendableModel(BaseModel):
    """Base model class for model classes that accept extension fields, i.e. with keys start with 'x-'"""

    model_config = pydantic.ConfigDict(
        # Allow extra properties, but don't keep them.
        extra='ignore',
    )


class PropertyPattern:
    def __init__(self, pattern: str) -> None:
        self.pattern = pattern


class ModelWithAdditionalProperties(BaseModel):
    model_config = pydantic.ConfigDict(
        extra='allow',
    )


class ModelWithPatternProperties(BaseModel):
    @pydantic.model_validator(mode='before')
    @classmethod
    def validate(cls, value: typing.Any, info: pydantic.ValidationInfo):
        for field_name, field_info in cls.model_fields.items():
            pattern_anno = find_annotation_optional(field_info.metadata, PropertyPattern)
            if not pattern_anno:
                continue

            pattern = re.compile(pattern_anno.pattern)

            pattern_props = {}
            for key, item in value.items():
                if not isinstance(key, str):
                    raise ValueError(key)
                if pattern.search(key):
                    pattern_props[key] = item
            for key in pattern_props:
                del value[key]
            value[field_name] = pattern_props

        return value


def validate_example_xor_examples(values: Mapping[str, typing.Any]) -> None:
    if 'examples' in values and 'example' in values:
        raise ValueError('Only either example or examples is allowed')



__all__ = (
    'Annotation',
    'ApiKeyAuth',
    'Auth',
    'AuthModule',
    'AuthorizationCodeOAuth2Flow',
    'BuiltinTypeHint',
    'ClientClass',
    'ClientCredentialsOAuth2Flow',
    'ClientInit',
    'ClientModule',
    'Field',
    'GenericTypeHint',
    'HttpBasicAuth',
    'HttpDigestAuth',
    'ImplicitOAuth2Flow',
    'MimeMap',
    'ModelType',
    'ModulePath',
    'NONE',
    'OperationFunction',
    'ParamLocation',
    'Parameter',
    'PasswordOAuth2Flow',
    'ResponseEnvelopeModel',
    'ResponseHeader',
    'ResponseMap',
    'SchemaClass',
    'SchemaModule',
    'SecurityRequirements',
    'TypeHint',
    'type_hint_or_union',
)

import dataclasses as dc
import itertools
from collections.abc import Iterable, MutableMapping, MutableSequence, MutableSet, Sequence
from functools import cached_property
from typing import Self

from ..openapi import ParameterLocation as ParamLocation
from .model import (
    Annotation,
    ApiKeyAuth,
    Auth,
    AuthorizationCodeOAuth2Flow,
    ClientClass,
    ClientCredentialsOAuth2Flow,
    ClientInit,
    Field,
    HttpBasicAuth,
    HttpDigestAuth,
    ImplicitOAuth2Flow,
    MimeMap,
    ModelType,
    OperationFunction,
    Parameter,
    PasswordOAuth2Flow,
    ResponseEnvelopeModel,
    ResponseHeader,
    ResponseMap,
    SchemaClass,
    SecurityRequirements,
)
from .module import AbstractModule, AuthModule, ClientModule, EmptyModule, ResponseEnvelopeModule, SchemaModule
from .module_path import ModulePath
from .type_hint import NONE, BuiltinTypeHint, GenericTypeHint, TypeHint, type_hint_or_union


@dc.dataclass
class ClientModel:
    client: ClientModule
    package: str
    schemas: MutableSequence[SchemaModule] = dc.field(default_factory=list)
    security_schemes: MutableMapping[str, Auth] = dc.field(default_factory=dict)
    _response_envelopes: MutableSequence[ResponseEnvelopeModule] = dc.field(default_factory=list)

    def packages(self: Self) -> Iterable[ModulePath]:
        # Used to create __init__.py files in otherwise empty packages

        known_packages: MutableSet[ModulePath] = {ModulePath(self.package)}

        for mod in itertools.chain(self.schemas, self._response_envelopes):
            path: ModulePath | None = mod.path
            while path := path.parent():
                if path in known_packages:
                    break
                yield path
                known_packages.add(path)

    @cached_property
    def modules(self) -> Sequence[AbstractModule]:
        return list(self._modules())

    def _modules(self) -> Iterable[AbstractModule]:
        known_modules = set()
        for mod in itertools.chain(self.schemas, self._response_envelopes):
            assert mod.path not in known_modules, mod.path
            known_modules.add(mod.path)
            yield mod

        for package in self.packages():
            if package not in known_modules:
                yield EmptyModule(path=package, body=None)

    def add_response_envelope_module(self, mod: ResponseEnvelopeModule):
        self._response_envelopes.append(mod)



import functools
import logging
from collections.abc import Callable
from typing import Concatenate

from . import openapi
from .stack import Stack

logger = logging.getLogger(__name__)


def resolve_ref[Target, R, **P](
    fn: Callable[Concatenate[Target, Stack, P], R],
) -> Callable[Concatenate[Target | openapi.Reference[Target], Stack, P], R]:
    @functools.wraps(fn)
    def wrapper(self, value: Target | openapi.Reference[Target], stack: Stack, *args) -> R:
        if isinstance(value, openapi.Reference):
            logger.debug('Resolving ref %s', value.ref)
            value, stack = self.resolve_ref(value)
        return fn(self, value, stack, *args)

    return wrapper



import dataclasses as dc
from collections.abc import Collection, Iterable, MutableSet
from typing import cast


@dc.dataclass(slots=True, frozen=True, kw_only=True)
class TypeHint:
    module: str
    name: str

    def __repr__(self):
        return self.full_name()

    def full_name(self):
        return self.module + ':' + self.name if self.module != 'builtins' else self.name

    @staticmethod
    def from_str(path: str) -> 'TypeHint':
        module, name = path.split(':')
        return TypeHint(module=module, name=name)

    @staticmethod
    def from_type(typ: type) -> 'TypeHint':
        if hasattr(typ, '__origin__'):
            raise ValueError('Generic types unsupported', typ)
        module = typ.__module__
        name = typ.__name__
        if module == 'builtins':
            return BuiltinTypeHint.from_str(name)
        else:
            return TypeHint(module=module, name=name)

    def imports(self) -> Iterable[str]:
        return [self.module]

    def is_union(self) -> bool:
        return False

    def is_none(self) -> bool:
        return False

    def dependencies(self) -> 'Iterable[TypeHint]':
        return ()

    def __eq__(self, other) -> bool:
        return isinstance(other, TypeHint) and self.module == other.module and self.name == other.name

    def __hash__(self) -> int:
        return self.module.__hash__() * 14159 + self.name.__hash__()


@dc.dataclass(slots=True, frozen=True, kw_only=True)
class BuiltinTypeHint(TypeHint):
    module: str = 'builtins'

    def __repr__(self):
        return self.full_name()

    @staticmethod
    def from_str(name: str) -> TypeHint:
        return BuiltinTypeHint(name=name)

    def full_name(self):
        return self.name


@dc.dataclass(slots=True, frozen=True, kw_only=True)
class GenericTypeHint(TypeHint):
    args: Iterable[TypeHint]

    @staticmethod
    def union_of(*types: TypeHint) -> 'GenericTypeHint':
        args: set[TypeHint] = set()
        for typ in types:
            if typ and typ.is_union():
                args.update(cast(GenericTypeHint, typ).args)
            else:
                args.add(typ)
        return GenericTypeHint(module='typing', name='Union', args=sorted(args, key=str))

    def is_union(self) -> bool:
        return self.module == 'typing' and self.name == 'Union'

    def dependencies(self) -> Iterable[TypeHint]:
        yield from self.args

    def __repr__(self) -> str:
        return self.full_name()

    def full_name(self) -> str:
        return f'{super(GenericTypeHint, self).full_name()}[{", ".join(arg.full_name() for arg in self.args)}]'

    @property
    def origin(self) -> TypeHint:
        return TypeHint(module=self.module, name=self.name)

    @staticmethod
    def list_of(item: TypeHint) -> 'GenericTypeHint':
        return GenericTypeHint(module='builtins', name='list', args=(item,))

    def __eq__(self, other) -> bool:
        return super(GenericTypeHint, self).__eq__(other) and (
            isinstance(other, GenericTypeHint) and self.args == other.args
        )

    def __hash__(self) -> int:
        hash_ = super(GenericTypeHint, self).__hash__()
        for arg in self.args:
            hash_ = (hash_ << 1) + arg.__hash__()
        return hash_


class NoneTypeHint(TypeHint):
    def __init__(self):
        super().__init__(module='types', name='NoneType')

    def full_name(self):
        return 'None'

    def imports(self) -> Iterable[str]:
        return ()

    def __eq__(self, other) -> bool:
        return isinstance(other, NoneTypeHint)

    def __hash__(self) -> int:
        return -1

    def is_none(self) -> bool:
        return True


NONE = NoneTypeHint()


def type_hint_or_union(types: Collection[TypeHint]) -> TypeHint:
    if not types:
        return NONE
    if len(types) == 1:
        return next(iter(types))
    else:
        return GenericTypeHint.union_of(*types)


def flatten(types: Iterable[TypeHint]) -> Iterable[TypeHint]:
    all_types = set()
    _flatten(types, all_types)
    return GenericTypeHint.union_of(*all_types).args


def _flatten(types: Iterable[TypeHint], target: MutableSet[TypeHint]) -> None:
    for typ in types:
        target.add(typ)
        _flatten(typ.dependencies(), target)



import importlib.metadata
import logging
import pathlib
from collections.abc import Mapping
from typing import Any, TextIO

import anyio
import asyncclick as click
import jinja2
import jinja2.loaders
import ruamel.yaml
import rybak

from .config import Config, load_config
from .load import document_handler_for, load_document
from .model import OpenApi30Converter, openapi, python

logger = logging.getLogger(__name__)


async def init_project(
    document_path: str,
    project_root: anyio.Path,
    package_name: str,
    save_document: bool,
) -> None:
    """Create project directory and pyproject file, download or copy the OpenAPI document"""

    if await project_root.exists():
        raise FileExistsError

    document_handler = document_handler_for(anyio.Path(), document_path)

    if save_document:
        document_root = anyio.Path('src/openapi')
        target_dir = project_root / document_root
        await target_dir.mkdir(parents=True)
        file_name = await document_handler.save_to(target_dir)
        config_document_path = str(document_root / file_name)
    else:
        logger.warning('Saving OpenAPI document is recommended for portable and repeatable builds')
        if not document_handler.is_url:
            document_path_path = pathlib.PurePath(document_path)
            if not document_path_path.is_absolute():
                # document path is a file outside of the project root.
                # If it's relative, make it relative to the project root.
                try:
                    config_document_path = document_path_path.relative_to(project_root, walk_up=True)
                except ValueError:
                    config_document_path = str(await anyio.Path.cwd() / document_path_path)
            else:
                config_document_path = document_path
        else:
            config_document_path = None

    config = Config(
        # if path is not URL and not saving the file
        document_path=config_document_path,
        origin=document_handler.path if document_handler.is_url else None,
        package=package_name,
    )

    yaml = ruamel.yaml.YAML(typ='safe')
    document = yaml.load(await document_handler.load())

    from rybak import TreeTemplate
    from rybak.jinja import JinjaAdapter

    TreeTemplate(
        JinjaAdapter(
            jinja2.Environment(
                loader=jinja2.loaders.PackageLoader('lapidary.render', package_path='templates/init'),
            )
        ),
        remove_suffixes=['.jinja'],
    ).render(
        dict(
            get_version=importlib.metadata.version,
            config=config.model_dump(exclude_unset=True, exclude_defaults=True, exclude_none=True),
            document=document,
        ),
        pathlib.Path(project_root),
    )


async def render_project(project_root: anyio.Path) -> None:
    config = await load_config(project_root)

    logger.info('Parse OpenAPI document')
    oa_doc = await load_document(project_root, config)
    model = prepare_python_model(oa_doc, config)

    logger.info('Render project')

    from rybak import TreeTemplate
    from rybak.jinja import JinjaAdapter

    template = TreeTemplate(
        JinjaAdapter(
            jinja2.Environment(
                loader=jinja2.loaders.PackageLoader('lapidary.render', package_path='templates/render'),
            )
        ),
        exclude_extend=[
            'includes',
        ],
        remove_suffixes=['.jinja'],
    )

    with RenderProgressBar(model) as event_sink:
        template.render(
            dict(
                model=model,
                get_version=importlib.metadata.version,
            ),
            pathlib.Path(project_root) / 'gen',
            event_sink=event_sink,
            remove_stale=True,
        )


class RenderProgressBar(rybak.EventSink):
    def __init__(self, model: python.ClientModel) -> None:
        self._progress_bar = click.progressbar(
            model.modules,
            label='Rendering modules',
            item_show_func=lambda item: item or '',
            show_pos=True,
        )

    def __enter__(self) -> Any:
        self._progress_bar.__enter__()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> bool | None:
        return self._progress_bar.__exit__(exc_type, exc_val, exc_tb)

    def writing_file(self, template: pathlib.PurePath, target: pathlib.Path) -> None:
        super().writing_file(template, target)
        if str(template) == '{{loop_over(model.modules).file_path}}.py.jinja':
            self._progress_bar.update(1, str(target).split('/')[-2])


async def dump_model(project_root: anyio.Path, process: bool, output: TextIO):
    from pprint import pprint

    config = await load_config(project_root)
    oa_doc = await load_document(project_root, config)

    if not process:
        yaml = ruamel.yaml.YAML(typ='safe')
        yaml.dump(oa_doc, output)

    else:
        py_model = prepare_python_model(oa_doc, config)
        pprint(py_model, output)


def prepare_python_model(oa_doc: Mapping, config: Config) -> python.ClientModel:
    oa_model = openapi.OpenApiModel.model_validate(oa_doc)
    with click.progressbar(
        length=len(oa_model.paths.paths),
        label='Processing paths',
        item_show_func=str,
        show_pos=True,
    ) as pbar:
        logger.info('Prepare python model')
        return OpenApi30Converter(
            python.ModulePath(config.package),
            oa_model,
            str(config.origin) if config.origin else None,
            path_progress=lambda item: pbar.update(1, item),
        ).process()



import keyword
import logging
import re
from typing import cast

import base62

logger = logging.getLogger(__name__)

VALID_IDENTIFIER_RE = re.compile(r'^[a-zA-Z]\w*$', re.ASCII)


def _escape_char(s: str) -> str:
    return f'u_{base62.encode(ord(s))}'


def escape_name(name: str) -> str:
    name = name.replace('u_', 'u' + _escape_char('_'))
    return re.sub('[^a-zA-Z]', lambda match: _escape_char(cast(str, match.group())), name[0]) + re.sub(
        '[^a-zA-Z0-9_]', lambda match: _escape_char(cast(str, match.group())), name[1:]
    )


def maybe_mangle_name(name: str) -> str:
    """
    Names that are Python keywords or in builtins get suffixed with an underscore (_).
    Names that are not valid Python identifiers or start with underscore are mangled by replacing invalid characters
    with u_{code}.
    Since 'u_' is the escaping prefix, it also gets replaced
    """
    if name is None or name == '':
        raise ValueError()

    if keyword.iskeyword(name):
        name = '\0' + name
    if not VALID_IDENTIFIER_RE.match(name) or 'u_' in name:
        return escape_name(name)
    else:
        return name



import logging
import sys
from pathlib import Path

import anyio
import asyncclick as click

logger = logging.getLogger(__name__)


@click.group()
@click.version_option(package_name='lapidary.render', prog_name='lapidary')
def app() -> None:
    pass


@app.command()
@click.argument('document')
@click.argument('project_root')
@click.argument('package_name')
@click.option('--save/--no-save', help='Copy the document in the project', default=False)
async def init(
    document: str,
    project_root: str,
    package_name: str,
    save: bool = False,
):
    """Initialize a new project.

    DOCUMENT: Path or URL of the OpenAPI document

    PROJECT_ROOT: Root directory of the generated project

    PACKAGE_NAME: Root package for the generated code
    """

    from .main import init_project

    try:
        await init_project(document, anyio.Path(project_root), package_name, save)
    except FileExistsError:
        raise click.ClickException('Target exists')


@app.command()
@click.argument('project_root', type=click.Path(exists=True, file_okay=False, dir_okay=True), default='.')
async def render(
    project_root: Path = Path(),
) -> None:
    """Generate Python code"""
    from .main import render_project

    await render_project(anyio.Path(project_root))


@app.command(hidden=True)
@click.argument('project_root', type=click.Path(exists=True, file_okay=False, dir_okay=True), default='.')
@click.option('--process', is_flag=True, help='Output processed python model', default=False)
async def dump_model(
    project_root: Path = Path(),
    process: bool = False,
) -> None:
    from .main import dump_model as dump_model_

    await dump_model_(anyio.Path(project_root), process, sys.stdout)



import dataclasses as dc
import enum
from collections.abc import Iterable, Mapping
from typing import Any, Literal, TypeAlias

from ..openapi import ParameterLocation as ParamLocation
from ..openapi import Style as ParamStyle
from .type_hint import NONE, TypeHint, type_hint_or_union

MimeType: TypeAlias = str
ResponseCode: TypeAlias = str
MimeMap: TypeAlias = Mapping[MimeType, TypeHint]
ResponseMap: TypeAlias = Mapping[ResponseCode, MimeMap]
SecurityRequirements: TypeAlias = Iterable[Mapping[str, Iterable[str]]]


@dc.dataclass
class Annotation:
    type: TypeHint
    field_props: dict[str, Any]
    style: str | None = None
    explode: bool | None = None
    allowReserved: bool | None = False


@dc.dataclass
class Field:
    name: str
    annotation: Annotation
    required: bool
    """
    Used for op method params. Required params are rendered before optional, and optional have default value None
    """

    deprecated: bool = False
    """Currently not used"""


@dc.dataclass
class Auth:
    name: str
    python_name: str
    type: str


@dc.dataclass(kw_only=True)
class ApiKeyAuth(Auth):
    key: str
    location: ParamLocation
    format: str
    type: str = 'api_key'


@dc.dataclass(kw_only=True)
class HttpAuth(Auth):
    scheme: str
    bearer_format: str | None


@dc.dataclass(kw_only=True)
class HttpBasicAuth(Auth):
    scheme: str = 'basic'
    type: str = 'http_basic'


@dc.dataclass(kw_only=True)
class HttpDigestAuth(Auth):
    scheme: str = 'digest'
    type: str = 'http_digest'


@dc.dataclass(kw_only=True)
class OpenIdConnectAuth(Auth):
    url: str
    type: str = 'openid_connect'


@dc.dataclass(kw_only=True)
class OAuth2AuthBase(Auth):
    scopes: dict[str, str]


@dc.dataclass(kw_only=True)
class ImplicitOAuth2Flow(OAuth2AuthBase):
    authorization_url: str
    type: str = 'oauth2_implicit'


@dc.dataclass(kw_only=True)
class PasswordOAuth2Flow(OAuth2AuthBase):
    token_url: str
    type: str = 'oauth2_password'


@dc.dataclass(kw_only=True)
class AuthorizationCodeOAuth2Flow(OAuth2AuthBase):
    authorization_url: str
    token_url: str
    type: str = 'oauth2_authorization_code'


@dc.dataclass(kw_only=True)
class ClientCredentialsOAuth2Flow(OAuth2AuthBase):
    token_url: str
    type: str = 'oauth2_client_credentials'


@dc.dataclass
class OperationFunction:
    name: str
    method: str
    path: str
    request_body: MimeMap
    params: Iterable['Parameter']
    responses: ResponseMap
    security: SecurityRequirements | None

    def dependencies(self) -> Iterable[TypeHint]:
        yield self.request_body_type
        for param in self.params:
            yield from param.dependencies()
        yield self.response_body_type

    @property
    def request_body_type(self) -> TypeHint:
        if not self.request_body:
            return NONE
        types = self.request_body.values()
        return type_hint_or_union(types)

    @property
    def response_body_type(self) -> TypeHint:
        types = set()
        for mime_map in self.responses.values():
            types.update(set(mime_map.values()))
        return type_hint_or_union(types)


@dc.dataclass(kw_only=True)
class Parameter:
    name: str
    """Python name"""

    alias: str | None
    """Header name"""

    type: TypeHint

    required: bool
    """
    Required params are rendered before optional, and optional have default value None
    """

    in_: ParamLocation
    default: Any = None
    """Default value, used only for global headers."""

    media_type: str | None = None
    style: ParamStyle | None
    explode: bool | None

    def dependencies(self) -> Iterable[TypeHint]:
        yield self.type


class ModelType(enum.Enum):
    model = 'python'
    exception = 'exception'


@dc.dataclass
class SchemaClass:
    class_name: str
    base_type: TypeHint

    allow_extra: bool = False
    docstr: str | None = None
    fields: list[Field] = dc.field(default_factory=list)
    model_type: ModelType = ModelType.model

    def dependencies(self) -> Iterable[TypeHint]:
        yield self.base_type
        for prop in self.fields:
            yield prop.annotation.type


@dc.dataclass
class ClientInit:
    default_auth: str | None = None
    auth_models: Mapping[str, Auth] = dc.field(default_factory=dict)
    base_url: str | None = None
    headers: Iterable[tuple[str, str]] = dc.field(default_factory=list)
    response_map: ResponseMap = dc.field(default_factory=dict)
    security: SecurityRequirements | None = None

    def dependencies(self) -> Iterable[TypeHint]:
        for mime_map in self.response_map.values():
            yield from mime_map.values()


@dc.dataclass(frozen=True)
class ClientClass:
    init_method: ClientInit
    methods: list[OperationFunction] = dc.field(default_factory=list)

    def dependencies(self) -> Iterable[TypeHint]:
        yield from self.init_method.dependencies()
        for fn in self.methods:
            yield from fn.dependencies()


@dc.dataclass(frozen=True)
class ResponseHeader:
    name: str
    alias: str
    type: TypeHint
    annotation: Literal['Cookie', 'Header', 'Link']

    def dependencies(self) -> Iterable[TypeHint]:
        yield self.type


@dc.dataclass(frozen=True)
class ResponseEnvelopeModel:
    name: str
    headers: Iterable[ResponseHeader]
    body_type: TypeHint

    def dependencies(self) -> Iterable[TypeHint]:
        yield self.body_type
        for header in self.headers:
            yield from header.dependencies()



import itertools
import logging
from collections import defaultdict
from collections.abc import Callable, Iterable, MutableMapping, Sequence

from .. import json_pointer, names
from . import openapi, python
from .refs import resolve_ref
from .stack import Stack

type ResolveRefFn[Target] = Callable[[openapi.Reference[Target]], tuple[Stack, Target]]

logger = logging.getLogger(__name__)


class OpenApi30SchemaConverter:
    def __init__(self, root_package: python.ModulePath, resolve_ref_: ResolveRefFn) -> None:
        self.root_package = root_package
        self.resolve_ref = staticmethod(resolve_ref_)
        self.schema_types: MutableMapping[Stack, tuple[python.SchemaClass, python.TypeHint]] = {}

    @resolve_ref
    def _process_schema_object(self, value: openapi.Schema, stack: Stack) -> python.TypeHint:
        if stack in self.schema_types:
            return self.schema_types[stack][1]

        name = value.lapidary_name or names.maybe_mangle_name(stack.top())
        base_type = (
            python.TypeHint.from_type(Exception)
            if value.lapidary_model_type is openapi.LapidaryModelType.exception
            else python.TypeHint.from_str('lapidary.runtime:ModelBase')
        )

        stack_props = stack.push('properties')
        fields = [
            self.process_property(prop_schema, stack_props.push(prop_name), prop_name, prop_name in value.required)
            for prop_name, prop_schema in value.properties.items()
        ]

        model_type = (
            python.ModelType[value.lapidary_model_type.name] if value.lapidary_model_type else python.ModelType.model
        )

        type_hint = resolve_type_hint(str(self.root_package), stack.push('schema', name))
        schema_class = python.SchemaClass(
            class_name=name,
            base_type=base_type,
            allow_extra=value.additional_properties is not False,
            fields=fields,
            docstr=value.description or None,
            model_type=model_type,
        )
        self.schema_types[stack] = schema_class, type_hint
        return type_hint

    @resolve_ref
    def process_property(self, value: openapi.Schema, stack: Stack, prop_name: str, required: bool) -> python.Field:
        name = names.maybe_mangle_name(prop_name)

        field_props = {FIELD_PROPS[k]: getattr(value, k) for k in value.model_fields_set if k in FIELD_PROPS}
        for k, v in field_props.items():
            if isinstance(v, str):
                if k == 'pattern':
                    field_props[k] = f"r'{v}'"
                else:
                    field_props[k] = f"'{v}'"

        if name != prop_name:
            field_props['alias'] = f"'{prop_name}'"

        typ = self.process_schema(value, stack)
        if value.nullable or not required or value.read_only or value.write_only:
            typ = python.GenericTypeHint.union_of(typ, python.NONE)

        return python.Field(
            name=name,
            annotation=python.Annotation(
                type=typ,
                field_props=field_props,
            ),
            required=required,
        )

    @resolve_ref
    def process_schema(self, value: openapi.Schema, stack: Stack, required: bool = True) -> python.TypeHint:
        assert isinstance(value, openapi.Schema)
        logger.debug('Process schema %s', stack)

        typ = self._process_schema(value, stack)

        if value.nullable or not required:
            typ = python.GenericTypeHint.union_of(typ, python.NONE)

        return typ

    def _get_one_of_type_hint(
        self,
        stack: Stack,
        one_of: Iterable[openapi.Reference[openapi.Schema] | openapi.Schema],
    ) -> python.TypeHint:
        return python.GenericTypeHint.union_of(
            *tuple(self.process_schema(sub_schema, stack.push(str(idx))) for idx, sub_schema in enumerate(one_of))
        )

    def _get_composite_type_hint(
        self,
        stack: Stack,
        schemas: list[openapi.Schema | openapi.Reference],
    ) -> python.TypeHint:
        if len(schemas) != 1:
            raise NotImplementedError(stack, 'Multiple component schemas (allOf, anyOf) are currently unsupported.')

        return self.process_schema(schemas[0], stack)

    def _process_schema(
        self,
        value: openapi.Schema,
        stack: Stack,
    ) -> python.TypeHint:
        if value.any_of:
            return self._get_composite_type_hint(stack.push('anyOf'), value.any_of)
        elif value.one_of:
            return self._get_one_of_type_hint(stack.push('oneOf'), value.one_of)
        elif value.all_of:
            return self._get_composite_type_hint(stack.push('allOf'), value.all_of)
        elif value.not_:
            raise NotImplementedError(stack.push('not'))
        elif value.type in PRIMITIVE_TYPES:
            return python.BuiltinTypeHint.from_str(PRIMITIVE_TYPES[value.type].__name__)
        elif value.type == openapi.Type.object:
            return self._process_schema_object(value, stack)
        elif value.type == openapi.Type.array:
            return python.GenericTypeHint.list_of(self.process_schema(value.items, stack.push('items')))
        elif value.type is None:
            return python.TypeHint.from_str('typing:Any')
        else:
            raise NotImplementedError(str(stack))

    @property
    def schema_modules(self) -> Iterable[python.SchemaModule]:
        modules: dict[python.ModulePath, list[python.SchemaClass]] = defaultdict(list)
        for pointer, schema_class_type in self.schema_types.items():
            schema_class, hint = schema_class_type
            modules[python.ModulePath(hint.module, is_module=True)].append(schema_class)
        return [
            python.SchemaModule(
                path=module,
                body=classes,
            )
            for module, classes in modules.items()
        ]


PRIMITIVE_TYPES = {
    openapi.Type.string: str,
    openapi.Type.integer: int,
    openapi.Type.number: float,
    openapi.Type.boolean: bool,
}

FIELD_PROPS = {
    'exclusive_maximum': 'lt',
    'exclusive_minimum': 'gt',
    'maximum': 'le',
    'max_items': 'max_length',
    'max_length': 'max_length',
    'max_properties': 'min_length',
    'minimum': 'ge',
    'min_items': 'min_length',
    'min_length': 'min_length',
    'min_properties': 'min_length',
    'multiple_of': 'multiple_of',
    'pattern': 'pattern',
}


def resolve_type_hint(root_package: str, pointer: str | Stack) -> python.TypeHint:
    if isinstance(pointer, Stack):
        parts: Sequence[str] = pointer.path[1:]
    else:
        parts = pointer.split('/')[1:]
    module_name = '.'.join(
        itertools.chain(
            (root_package,), [names.maybe_mangle_name(json_pointer.decode_json_pointer(part)) for part in parts[:-1]]
        )
    )
    top = names.maybe_mangle_name(parts[-1])
    return python.TypeHint(module=module_name, name=top)



def encode_json_pointer(path: str) -> str:
    return path.replace('~', '~0').replace('/', '~1')


def decode_json_pointer(encoded: str) -> str:
    return encoded.replace('~1', '/').replace('~0', '~')



import abc
import dataclasses as dc
from collections.abc import Iterable, Mapping

from .model import ClientClass, ResponseEnvelopeModel, SchemaClass
from .module_path import ModulePath
from .type_hint import NONE, TypeHint, flatten

template_imports = [
    'builtins',
    'typing',
    'typing_extensions',
    'lapidary.runtime',
]


@dc.dataclass(frozen=True, kw_only=True)
class AbstractModule[Body](abc.ABC):
    path: ModulePath = dc.field()
    module_type: str
    body: Body = dc.field()

    @abc.abstractmethod
    def dependencies(self) -> Iterable[TypeHint]:
        pass

    @property
    def imports(self) -> Iterable[str]:
        dependencies = flatten(self.dependencies())
        return sorted(
            {
                dep.module
                for dep in dependencies
                if dep.module not in template_imports and dep != NONE and dep.module != str(self.path)
            }
        )

    @property
    def file_path(self) -> str:
        return self.path.to_path()


@dc.dataclass(frozen=True, kw_only=True)
class AuthModule(AbstractModule[Mapping[str, TypeHint]]):
    module_type = 'auth'

    @property
    def dependencies(self) -> Iterable[TypeHint]:
        return self.body.values()


@dc.dataclass(frozen=True, kw_only=True)
class ClientModule(AbstractModule[ClientClass]):
    module_type: str = dc.field(default='client')

    def dependencies(self) -> Iterable[TypeHint]:
        return self.body.dependencies()


@dc.dataclass(frozen=True, kw_only=True)
class EmptyModule(AbstractModule[None]):
    """Module used to generate empty __init__.py files"""

    module_type: str = dc.field(default='empty')
    body: None = None

    def dependencies(self) -> Iterable[TypeHint]:
        return ()


@dc.dataclass(frozen=True, kw_only=True)
class ResponseEnvelopeModule(AbstractModule[ResponseEnvelopeModel]):
    module_type: str = dc.field(default='response')

    def dependencies(self) -> Iterable[TypeHint]:
        return self.body.dependencies()


@dc.dataclass(frozen=True, kw_only=True)
class SchemaModule(AbstractModule[Iterable[SchemaClass]]):
    """
    One schema module per schema element directly under #/components/schemas, containing that schema and all non-reference schemas.
    One schema module for inline request and for response body for each operation
    """

    module_type: str = 'schema'

    def dependencies(self) -> Iterable[TypeHint]:
        for schema in self.body:
            yield from schema.dependencies()



import typing

T = typing.TypeVar('T')


def find_annotation_optional(
    metadata: list,
    annotation_type: type[T],
) -> T | None:
    annos = [anno for anno in metadata if isinstance(anno, annotation_type)]
    if len(annos) > 1:
        raise TypeError(f'Multiple {annotation_type} annotations')
    return annos[0] if annos else None
